# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

GDAL_TO_NUMPY = {1: "int8",
                 2: "uint16",
                 3: "int16",
                 4: "uint32",
                 5: "int32",
                 6: "float32",
                 7: "float64",
                 10: "complex64",
                 11: "complex128"}



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
import os
import uuid
from tempfile import mkstemp, gettempdir


def _copy_to_file(raster, out_file):
    """

    """
    try:
        raster._gdal_driver.CopyFiles(out_file, raster._file)
        # out_ds = raster._gdal_driver.CreateCopy(out_file, raster._gdal_dataset, strict=0)
        # out_ds = None
        return 0
    except RuntimeError:
        return 1


class File:

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TempFile(File):

    def __del__(self):
        try:
            os.remove(self.path)
        except FileNotFoundError:
            pass


class NamedTempFile(TempFile):

    def __init__(self, extension):
        self.name = os.path.join(gettempdir(), str(uuid.uuid4()))
        super().__init__(self.name + "." + extension)


class ShapeTempFile(NamedTempFile):

    def __init__(self):
        super().__init__("shp")

    def __del__(self):
        super().__del__()
        for ext in [".shx", ".dbf", ".prj", ".cpg"]:
            try:
                os.remove(self.name + ext)
            except FileNotFoundError:
                pass


class RasterTempFile(TempFile):
    """ Create temporary raster file

    """
    def __init__(self, extension):
        super().__init__(mkstemp(suffix='.' + extension)[1])


class VrtTempFile(TempFile):

    def __init__(self):
        super().__init__(mkstemp(suffix='.vrt')[1])



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
import numpy as np

from pyrasta.tools import _gdal_temp_dataset, _return_raster
from pyrasta.tools.mapping import GDAL_TO_NUMPY
from pyrasta.tools.windows import get_block_windows, get_xy_block_windows
from pyrasta.utils import split_into_chunks
from tqdm import tqdm

import gdal

OP_WINDOW_SIZE = 1000


@_return_raster
def _op(raster1, out_file, raster2, op_type):
    """ Basic arithmetic operations

    """
    out_ds = _gdal_temp_dataset(out_file, raster1._gdal_driver,
                                raster1._gdal_dataset.GetProjection(), raster1.x_size,
                                raster1.y_size, raster1.nb_band, raster1.geo_transform,
                                gdal.GetDataTypeByName('Float32'), raster1.no_data)

    for band in range(1, raster1.nb_band + 1):

        for window in get_block_windows(OP_WINDOW_SIZE, raster1.x_size, raster1.y_size):
            arrays = []
            for src in [raster1, raster2]:
                try:
                    arrays.append(src._gdal_dataset.GetRasterBand(
                        band).ReadAsArray(*window).astype("float32"))
                except AttributeError:
                    arrays.append(src)
            # array1 = raster1._gdal_dataset.GetRasterBand(
            #     band).ReadAsArray(*window).astype("float32")
            # try:
            #     array2 = raster2._gdal_dataset.GetRasterBand(
            #         band).ReadAsArray(*window).astype("float32")
            # except AttributeError:
            #     array2 = raster2  # If second input is not a raster but a scalar

            if op_type == "add":
                result = arrays[0] + arrays[1]
            elif op_type == "sub":
                result = arrays[0] - arrays[1]
            elif op_type == "rsub":
                result = arrays[1] - arrays[0]
            elif op_type == "mul":
                result = arrays[0] * arrays[1]
            elif op_type == "pow":
                result = arrays[0] ** arrays[1]
            elif op_type == "rpow":
                result = arrays[1] ** arrays[0]
            elif op_type == "truediv":
                result = arrays[0] / arrays[1]
            elif op_type == "rtruediv":
                return arrays[1] / arrays[0]
            else:
                result = None

            out_ds.GetRasterBand(band).WriteArray(result, window[0], window[1])

    # Close dataset
    out_ds = None


@_return_raster
def _raster_calculation(raster_class, out_file, gdal_driver, sources,
                        fhandle, window_size, input_type, output_type,
                        no_data, nb_processes, chunksize, description):
    """ Calculate raster expression

    """
    if not hasattr(window_size, "__getitem__"):
        window_size = (window_size, window_size)

    master_raster = sources[0]
    window_gen = ([src._gdal_dataset.ReadAsArray(*w).astype(GDAL_TO_NUMPY[input_type]) for src in
                   sources] for w in get_xy_block_windows(window_size,
                                                          master_raster.x_size,
                                                          master_raster.y_size))
    width = int(master_raster.x_size /
                window_size[0]) + min(1, master_raster.x_size % window_size[0])
    height = int(master_raster.y_size /
                 window_size[1]) + min(1, master_raster.y_size % window_size[1])

    # Initialization
    is_first_run = True
    y = 0

    if description:
        iterator = tqdm(split_into_chunks(window_gen, width),
                        total=height,
                        desc=description)
    else:
        iterator = split_into_chunks(window_gen, width)

    for win_gen in iterator:

        with mp.Pool(processes=nb_processes) as pool:
            list_of_arrays = list(pool.map(fhandle,
                                           win_gen,
                                           chunksize=chunksize))

        result = np.concatenate(list_of_arrays, axis=list_of_arrays[0].ndim - 1)

        if is_first_run:
            if result.ndim == 2:
                nb_band = 1
            else:
                nb_band = result.shape[0]

            out_ds = _gdal_temp_dataset(out_file,
                                        gdal_driver,
                                        master_raster._gdal_dataset.GetProjection(),
                                        master_raster.x_size,
                                        master_raster.y_size, nb_band,
                                        master_raster.geo_transform,
                                        output_type,
                                        no_data)

            is_first_run = False

        if nb_band == 1:
            out_ds.GetRasterBand(1).WriteArray(result, 0, y)
        else:
            for band in range(nb_band):
                out_ds.GetRasterBand(band + 1).WriteArray(result[band, :, :],
                                                          0, y)

        y += window_size[1]

    # Close dataset
    out_ds = None



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from functools import partial

from pyrasta.io_ import ESRI_DRIVER
from pyrasta.io_.files import ShapeTempFile, _copy_to_file, NamedTempFile

import gdal


def _mask(arrays, no_data):
    src = arrays[0]
    mask = arrays[1]
    src[mask == 1] = no_data

    return src


def _raster_mask(raster, geodataframe, driver, output_type, no_data, all_touched,
                 window_size):
    """ Apply mask into raster

    """
    mask = raster.__class__.rasterize(geodataframe,
                                      raster.crs.to_wkt(),
                                      raster.x_size,
                                      raster.y_size,
                                      raster.geo_transform,
                                      burn_values=[1],
                                      all_touched=all_touched)

    return raster.__class__.raster_calculation([raster, mask],
                                               partial(_mask, no_data=no_data),
                                               gdal_driver=driver,
                                               data_type=output_type,
                                               no_data=no_data,
                                               description="Compute mask",
                                               window_size=window_size,
                                               nb_processes=1,
                                               chunksize=1)


    # with ShapeTempFile() as shp_file, \
    #         NamedTempFile(raster._gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:
    #
    #     _copy_to_file(raster, out_file.path)
    #     geodataframe.to_file(shp_file.path, driver=ESRI_DRIVER)
    #     out_ds = gdal.Open(out_file.path, 1)
    #     gdal.Rasterize(out_ds,
    #                    shp_file.path,
    #                    bands=[bd + 1 for bd in range(raster.nb_band)],
    #                    burnValues=[10],
    #                    allTouched=all_touched)
    #
    # out_ds = None

    # return raster.__class__(out_file.path)
    # new_raster = raster.__class__(out_file.path)
    # new_raster._temp_file = out_file

    # return new_raster



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
from functools import partial
from itertools import tee

import numpy as np
from tqdm import tqdm


STATISTIC_FUNC = dict(median=np.median,
                      mean=np.mean,
                      min=np.min,
                      max=np.max,
                      sum=np.sum)


def _histogram(raster, nb_bins, normalized):
    """ Compute histogram of raster values

    """
    histogram = []

    for band in range(raster.nb_band):
        edges = np.linspace(raster.min, raster.max, nb_bins + 1)
        hist_x = edges[0:-1] + (edges[1::] - edges[0:-1])/2
        hist_y = np.asarray(
            raster._gdal_dataset.GetRasterBand(band + 1).GetHistogram(min=raster.min[band],
                                                                      max=raster.max[band],
                                                                      buckets=nb_bins))
        if normalized:
            hist_y = hist_y / np.sum(hist_y)

        histogram.append((hist_x, hist_y))

    return histogram


def _zonal_stats(raster, layer, band, stats, customized_stat,
                 no_data, all_touched, show_progressbar, nb_processes):
    """ Retrieve zonal statistics from raster corresponding to features in layer
    
    Parameters
    ----------
    raster: RasterBase
        Raster from which zonal statistics must be computed
    layer: geopandas.GeoDataFrame or gistools.layer.GeoLayer
        Geographic layer as a GeoDataFrame or GeoLayer
    band: int
        band number
    stats: list[str]
        list of strings of valid available statistics:
        - 'mean' returns average over the values within each zone
        - 'median' returns median
        - 'sum' returns the sum of all values in zone
        - 'min' returns minimum value
        - 'max' returns maximum value
    customized_stat: dict
        User's own customized statistic function
        as {'your_function_name': function}
    no_data: int or float
        No data value
    all_touched: bool
        Whether to include every raster cell touched by a geometry, or only
        those having a center point within the polygon.
    show_progressbar: bool
        if True, show progress bar status
    nb_processes: int
        Number of parallel processes

    Returns
    -------

    """
    def zone_gen(ras, bds, bd=1):
        for boundary in bds:
            try:
                yield ras.read_array(bd, boundary)
            except RuntimeError:
                yield None

    stats_calc = {name: STATISTIC_FUNC[name] for name in stats}
    if customized_stat is not None:
        stats_calc.update(customized_stat)

    layer["ID"] = layer.index
    raster_layer = raster.rasterize(layer, raster.projection, raster.x_size,
                                    raster.y_size, raster.geo_transform,
                                    attribute="ID", all_touched=all_touched)

    bounds = layer.bounds.to_numpy()
    zone = zone_gen(raster, bounds, band)
    zone_id = zone_gen(raster_layer, bounds)
    multi_gen = tee(zip(layer.index, zone, zone_id), len(stats_calc))

    iterator = zip(multi_gen, stats_calc.keys())

    output = dict()
    with mp.Pool(processes=nb_processes) as pool:
        if show_progressbar:
            for generator, name in iterator:
                output[name] = list(tqdm(pool.starmap(partial(_compute_stat_in_feature,
                                                              no_data=raster.no_data,
                                                              stat_function=stats_calc[name]),
                                                      generator),
                                         total=len(layer),
                                         unit_scale=True,
                                         desc=f"Compute zonal {name}"))
        else:
            for generator, name in iterator:
                output[name] = list(pool.starmap(partial(_compute_stat_in_feature,
                                                         no_data=raster.no_data,
                                                         stat_function=stats_calc[name]),
                                                 generator))

    return output


def _compute_stat_in_feature(idx, zone, zone_id, no_data, stat_function):

    if zone is not None and zone_id is not None:
        if np.isnan(no_data):
            values = zone[(zone_id == idx) & ~np.isnan(zone)]
        else:
            values = zone[(zone_id == idx) & (zone != no_data)]

        if values.size != 0:
            return stat_function(values)
        else:
            return np.nan
    else:
        return np.nan



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from pyrasta.io_ import ESRI_DRIVER
from pyrasta.io_.files import RasterTempFile, ShapeTempFile
from pyrasta.tools import _return_raster, _gdal_temp_dataset

import gdal


@_return_raster
def _clip_raster_by_extent(raster, out_file, bounds, no_data):
    """ Clip raster by extent

    Parameters
    ----------
    raster: pyrasta.raster.RasterBase
    out_file: pyrasta.io_.files.RasterTempFile
    bounds: tuple
        boundaries as (minx, miny, maxx, maxy)
    no_data: int or float
        No data value

    Returns
    -------
    RasterBase

    """

    minx = max(bounds[0], raster.bounds[0])
    miny = max(bounds[1], raster.bounds[1])
    maxx = min(bounds[2], raster.bounds[2])
    maxy = min(bounds[3], raster.bounds[3])

    if minx >= maxx or miny >= maxy:
        raise ValueError("requested extent out of raster boundaries")

    gdal.Warp(out_file,
              raster._gdal_dataset,
              outputBounds=bounds,
              srcNodata=raster.no_data,
              dstNodata=no_data,
              outputType=raster.data_type)


def _clip_raster_by_mask(raster, geodataframe, no_data, all_touched):
    """ Clip raster by mask from geographic layer

    Parameters
    ----------
    raster: pyrasta.raster.RasterBase
        raster to clip
    geodataframe: geopandas.GeoDataFrame or gistools.layer.GeoLayer
    no_data: float or int
        No data value
    all_touched: bool
        if True, clip all pixels that are touched, otherwise clip
        if pixel's centroids are within boundaries

    Returns
    -------
    RasterBase

    """
    clip_raster = raster.clip(bounds=geodataframe.total_bounds)

    with ShapeTempFile() as shp_file, \
            RasterTempFile(clip_raster._gdal_driver.GetMetadata()['DMD_EXTENSION']) as r_file:

        geodataframe.to_file(shp_file.path, driver=ESRI_DRIVER)

        out_ds = _gdal_temp_dataset(r_file.path,
                                    clip_raster._gdal_driver,
                                    clip_raster._gdal_dataset.GetProjection(),
                                    clip_raster.x_size,
                                    clip_raster.y_size,
                                    clip_raster.nb_band,
                                    clip_raster.geo_transform,
                                    clip_raster.data_type,
                                    clip_raster.no_data)

        gdal.Rasterize(out_ds,
                       shp_file.path,
                       burnValues=[1],
                       allTouched=all_touched)

    out_ds = None

    return clip_raster.__class__.raster_calculation([clip_raster,
                                                     clip_raster.__class__(r_file.path)],
                                                    lambda x, y: x*y,
                                                    no_data=no_data,
                                                    showprogressbar=False)



from setuptools import setup, find_packages

import pyrasta

with open("README.md", 'r') as fh:
    long_description = fh.read()

with open("requirements.txt") as req:
    install_req = req.read().splitlines()

setup(name='pyrasta',
      version=pyrasta.__version__,
      description='Some tools for fast and easy raster processing',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://framagit.org/benjaminpillot/pyrasta',
      author='Benjamin Pillot',
      author_email='benjaminpillot@riseup.net',
      install_requires=install_req,
      python_requires='>=3',
      license='MIT',
      packages=find_packages(exclude="pyrasta/algorithms"),
      zip_safe=False)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from pyrasta.io_ import ESRI_DRIVER
from pyrasta.io_.files import ShapeTempFile
from pyrasta.tools import _gdal_temp_dataset, _return_raster

import gdal


@_return_raster
def _rasterize(raster_class, out_file, gdal_driver, geodataframe,
               burn_values, attribute, projection, x_size, y_size,
               nb_band, geo_transform, data_type, no_data, all_touched):
    """ Rasterize geographic layer

    Parameters
    ----------
    raster_class: RasterBase
        Raster class to return
    out_file: str
        Output file to which raster is written
    gdal_driver: gdal.Driver
        GDAL driver
    geodataframe: geopandas.GeoDataFrame or gistools.layer.GeoLayer
        Geographic layer to be rasterized
    burn_values: None or list[float] or list[int]
        list of values to burn in each band, excusive with attribute
    attribute: str
        attribute in layer from which burn value must be retrieved
    projection
    x_size: int
    y_size: int
    nb_band: int
    geo_transform: tuple
    data_type
    no_data
    all_touched: bool

    Returns
    -------

    """

    with ShapeTempFile() as shp_file:

        geodataframe.to_file(shp_file.path, driver=ESRI_DRIVER)

        out_ds = _gdal_temp_dataset(out_file,
                                    gdal_driver,
                                    projection,
                                    x_size,
                                    y_size,
                                    nb_band,
                                    geo_transform,
                                    data_type,
                                    no_data)

        gdal.Rasterize(out_ds,
                       shp_file.path,
                       bands=[bd + 1 for bd in range(nb_band)],
                       burnValues=burn_values,
                       attribute=attribute,
                       allTouched=all_touched)

    out_ds = None

    # Be careful with the temp file, make a pointer to be sure
    # the Python garbage collector does not destroy it !
    # raster = raster_class(out_file.path)
    # raster._temp_file = out_file
    #
    # return raster



# -*- coding: utf-8 -*-

""" Functions related to CRS conversion and computation

More detailed description.
"""

import osr
import pyproj


def is_equal_proj(proj1, proj2):
    """ Compare 2 projections

    Parameters
    ----------
    proj1: int or str or dict or pyproj.Proj
        valid projection name
    proj2: int or str or dict or pyproj.Proj
        valid projection name

    Returns
    -------
    boolean:
        True or False
    """
    # From an idea from https://github.com/jswhit/pyproj/issues/15
    # Use OGR library to compare projections
    srs = [srs_from(proj1), srs_from(proj2)]

    return bool(srs[0].IsSame(srs[1]))


def proj4_from_wkt(wkt):
    """ Convert wkt srs to proj4

    Description
    -----------

    Parameters
    ----------
    wkt:

    Returns
    -------

    """
    srs = osr.SpatialReference()
    srs.ImportFromWkt(wkt)

    return srs.ExportToProj4()


def proj4_from(proj):
    """ Convert projection to proj4 string

    Description
    -----------
    Convert projection string, dictionary, etc.
    to proj4 string

    Parameters
    ----------
    proj:

    Returns
    -------

    """
    if type(proj) == int:
        try:
            proj4_str = pyproj.Proj('epsg:%d' % proj).srs
        except (ValueError, RuntimeError):
            raise ValueError("Invalid EPSG code")
    elif type(proj) == str or type(proj) == dict:
        try:
            proj4_str = pyproj.Proj(proj).srs
        except RuntimeError:
            try:
                proj4_str = proj4_from_wkt(proj)
            except (RuntimeError, TypeError):
                raise ValueError("Invalid projection string or dictionary")
    elif type(proj) == pyproj.Proj:
        proj4_str = proj.srs
    else:
        raise ValueError("Invalid projection format: '{}'".format(type(proj)))

    return proj4_str


def srs_from(proj):
    """ Get spatial reference system from projection

    Description
    -----------

    Parameters
    ----------
    proj:

    Returns
    -------
    SpatialReference instance (osgeo.osr package)
    """
    proj4 = proj4_from(proj)
    srs = osr.SpatialReference()
    srs.ImportFromProj4(proj4)

    return srs


def wkt_from(proj):
    """ Get WKT spatial reference system from projection

    Description
    -----------

    Parameters
    ----------
    proj:

    Returns
    -------
    """
    return srs_from(proj).ExportToWkt()
